# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import itertools

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list

content = ak.contents.NumpyArray(
    np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
)
offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
listoffsetarray = ak.contents.ListOffsetArray(offsets, content)
regulararray = ak.contents.RegularArray(listoffsetarray, 2, zeros_length=0)
starts = ak.index.Index64(np.array([0, 1]))
stops = ak.index.Index64(np.array([2, 3]))
listarray = ak.contents.ListArray(starts, stops, regulararray)


def test_simple_type():
    assert str(ak.operations.type(content)) == "10 * float64"


def test_type():
    assert str(ak.operations.type(regulararray)) == "3 * 2 * var * float64"


def test_iteration():
    assert to_list(regulararray) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]


def test_getitem_at():
    assert to_list(regulararray[0]) == [[0.0, 1.1, 2.2], []]
    assert to_list(regulararray[1]) == [[3.3, 4.4], [5.5]]
    assert to_list(regulararray[2]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert regulararray.to_typetracer()[2].form == regulararray[2].form


def test_getitem_range():
    assert to_list(regulararray[1:]) == [
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert regulararray.to_typetracer()[1:].form == regulararray[1:].form
    assert to_list(regulararray[:-1]) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]
    assert regulararray.to_typetracer()[:-1].form == regulararray[:-1].form


def test_getitem():
    assert to_list(regulararray[(0,)]) == [[0.0, 1.1, 2.2], []]
    assert to_list(regulararray[(1,)]) == [[3.3, 4.4], [5.5]]
    assert to_list(regulararray[(2,)]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert regulararray.to_typetracer()[(2,)].form == regulararray[(2,)].form
    assert to_list(regulararray[(slice(1, None, None),)]) == [
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        regulararray.to_typetracer()[(slice(1, None, None),)].form
        == regulararray[(slice(1, None, None),)].form
    )
    assert to_list(regulararray[(slice(None, -1, None),)]) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
    ]
    assert (
        regulararray.to_typetracer()[(slice(None, -1, None),)].form
        == regulararray[(slice(None, -1, None),)].form
    )


def test_getitem_deeper():
    assert to_list(listarray) == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]

    assert to_list(listarray[0, 0, 0]) == [0.0, 1.1, 2.2]
    assert to_list(listarray[0, 0, 1]) == []
    assert to_list(listarray[0, 1, 0]) == [3.3, 4.4]
    assert to_list(listarray[0, 1, 1]) == [5.5]
    assert to_list(listarray[1, 0, 0]) == [3.3, 4.4]
    assert to_list(listarray[1, 0, 1]) == [5.5]
    assert to_list(listarray[1, 1, 0]) == [6.6, 7.7, 8.8, 9.9]
    assert to_list(listarray[1, 1, 1]) == []
    assert listarray.to_typetracer()[1, 1, 1].form == listarray[1, 1, 1].form

    assert to_list(listarray[0, 0, 0:]) == [[0.0, 1.1, 2.2], []]
    assert to_list(listarray[0, 0, 1:]) == [[]]
    assert to_list(listarray[0, 1, 0:]) == [[3.3, 4.4], [5.5]]
    assert to_list(listarray[0, 1, 1:]) == [[5.5]]
    assert to_list(listarray[1, 0, 0:]) == [[3.3, 4.4], [5.5]]
    assert to_list(listarray[1, 0, 1:]) == [[5.5]]
    assert to_list(listarray[1, 1, 0:]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert to_list(listarray[1, 1, 1:]) == [[]]
    assert listarray.to_typetracer()[1, 1, 1:].form == listarray[1, 1, 1:].form

    assert to_list(listarray[[1], 0, 0:]) == [[[3.3, 4.4], [5.5]]]
    assert listarray.to_typetracer()[[1], 0, 0:].form == listarray[[1], 0, 0:].form
    assert to_list(listarray[[1, 0], 0, 0:]) == [
        [[3.3, 4.4], [5.5]],
        [[0.0, 1.1, 2.2], []],
    ]
    assert (
        listarray.to_typetracer()[[1, 0], 0, 0:].form == listarray[[1, 0], 0, 0:].form
    )

    assert to_list(listarray[:, :, [0, 1]]) == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]
    assert to_list(listarray[:, :, [1, 0]]) == [
        [[[], [0.0, 1.1, 2.2]], [[5.5], [3.3, 4.4]]],
        [[[5.5], [3.3, 4.4]], [[], [6.6, 7.7, 8.8, 9.9]]],
    ]
    assert to_list(listarray[:, :, [1, 0, 1]]) == [
        [[[], [0.0, 1.1, 2.2], []], [[5.5], [3.3, 4.4], [5.5]]],
        [[[5.5], [3.3, 4.4], [5.5]], [[], [6.6, 7.7, 8.8, 9.9], []]],
    ]
    assert to_list(listarray[:, :2, [0, 1]]) == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]
    assert to_list(listarray[:1, [0, 0, 1, 1], [0, 1, 0, 1]]) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5]]
    ]
    assert to_list(listarray[:1, [1, 1, 0, 0], [1, 0, 1, 0]]) == [
        [[5.5], [3.3, 4.4], [], [0.0, 1.1, 2.2]]
    ]
    assert (
        listarray[:1, [1, 1, 0, 0], [1, 0, 1, 0]].form
        == listarray.to_typetracer()[:1, [1, 1, 0, 0], [1, 0, 1, 0]].form
    )


content2 = ak.contents.NumpyArray(np.arange(2 * 3 * 5 * 7).reshape(-1, 7))
regulararrayA = ak.contents.RegularArray(content2, 5, zeros_length=0)
regulararrayB = ak.contents.RegularArray(regulararrayA, 3, zeros_length=0)
modelA = np.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7)
modelB = np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7)


def test_numpy():
    assert to_list(regulararrayA) == to_list(modelA)
    assert to_list(regulararrayB) == to_list(modelB)

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert to_list(modelA[cuts]) == to_list(regulararrayA[cuts])
            if depth < 3:
                assert (
                    regulararrayA.to_typetracer()[cuts].form == regulararrayA[cuts].form
                )

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            (slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth
        ):
            assert to_list(modelA[cuts]) == to_list(regulararrayA[cuts])
            if depth < 3:
                assert (
                    regulararrayA.to_typetracer()[cuts].form == regulararrayA[cuts].form
                )

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            (slice(1, None), slice(None, -1), 2, -2), depth
        ):
            assert to_list(modelA[cuts]) == to_list(regulararrayA[cuts])
            if depth < 3:
                assert (
                    regulararrayA.to_typetracer()[cuts].form == regulararrayA[cuts].form
                )

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            ([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth
        ):
            assert to_list(modelA[cuts]) == to_list(regulararrayA[cuts])
            if depth < 3:
                assert (
                    regulararrayA.to_typetracer()[cuts].form == regulararrayA[cuts].form
                )

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            ([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth
        ):
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert to_list(modelA[cuts]) == to_list(regulararrayA[cuts])
            if depth < 3:
                assert (
                    regulararrayA.to_typetracer()[cuts].form == regulararrayA[cuts].form
                )

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert to_list(modelB[cuts]) == to_list(regulararrayB[cuts])
            if depth < 4:
                assert (
                    regulararrayB.to_typetracer()[cuts].form == regulararrayB[cuts].form
                )

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations(
            (-1, 0, 1, slice(1, None), slice(None, -1)), depth
        ):
            assert to_list(modelB[cuts]) == to_list(regulararrayB[cuts])
            if depth < 4:
                assert (
                    regulararrayB.to_typetracer()[cuts].form == regulararrayB[cuts].form
                )

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations(
            (-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth
        ):
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert to_list(modelB[cuts]) == to_list(regulararrayB[cuts])
            if depth < 4:
                assert (
                    regulararrayB.to_typetracer()[cuts].form == regulararrayB[cuts].form
                )


def test_maybe_to_Numpy():
    array = ak.highlevel.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    ).layout
    array2 = ak.highlevel.Array([3, 6, None, None, -2, 6], check_valid=True).layout
    assert to_list(array[array2]) == [
        3.3,
        6.6,
        None,
        None,
        8.8,
        6.6,
    ]
    assert array.to_typetracer()[array2].form == array[array2].form

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    regulararray = ak.contents.RegularArray(content, 4, zeros_length=0)

    array3 = ak.highlevel.Array([2, 1, 1, None, -1], check_valid=True).layout
    numpyarray = regulararray.maybe_to_NumpyArray()
    assert to_list(numpyarray[array3]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert numpyarray.to_typetracer()[array3].form == numpyarray[array3].form
    assert to_list(numpyarray[:, array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]

    a = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(
            np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
        ),
        3,
    )
    assert len(a) == 2
    a = a.maybe_to_NumpyArray()
    assert isinstance(
        a[1,],
        ak.contents.numpyarray.NumpyArray,
    )



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_BitMaskedArray_to_ByteMaskedArray_1():
    tobytemask = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    frombitmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = True
    lsb_order = True
    funcC = cupy_backend['awkward_BitMaskedArray_to_ByteMaskedArray', cupy.int8, cupy.uint8]
    funcC(tobytemask, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tobytemask = [False, True, True, True, True, True, True, True, False, True, True, True, True, True, True, True, False, True, True, True, True, True, True, True]
    assert cupy.array_equal(tobytemask[:len(pytest_tobytemask)], cupy.array(pytest_tobytemask))

def test_cuda_awkward_BitMaskedArray_to_ByteMaskedArray_2():
    tobytemask = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    frombitmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = False
    lsb_order = False
    funcC = cupy_backend['awkward_BitMaskedArray_to_ByteMaskedArray', cupy.int8, cupy.uint8]
    funcC(tobytemask, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tobytemask = [False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    assert cupy.array_equal(tobytemask[:len(pytest_tobytemask)], cupy.array(pytest_tobytemask))

def test_cuda_awkward_BitMaskedArray_to_ByteMaskedArray_3():
    tobytemask = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    frombitmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = True
    lsb_order = False
    funcC = cupy_backend['awkward_BitMaskedArray_to_ByteMaskedArray', cupy.int8, cupy.uint8]
    funcC(tobytemask, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tobytemask = [True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True]
    assert cupy.array_equal(tobytemask[:len(pytest_tobytemask)], cupy.array(pytest_tobytemask))

def test_cuda_awkward_BitMaskedArray_to_ByteMaskedArray_4():
    tobytemask = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    frombitmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = True
    lsb_order = True
    funcC = cupy_backend['awkward_BitMaskedArray_to_ByteMaskedArray', cupy.int8, cupy.uint8]
    funcC(tobytemask, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tobytemask = [False, True, True, True, True, True, True, True, False, True, True, True, True, True, True, True, False, True, True, True, True, True, True, True]
    assert cupy.array_equal(tobytemask[:len(pytest_tobytemask)], cupy.array(pytest_tobytemask))

def test_cuda_awkward_BitMaskedArray_to_ByteMaskedArray_5():
    tobytemask = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    frombitmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = False
    lsb_order = False
    funcC = cupy_backend['awkward_BitMaskedArray_to_ByteMaskedArray', cupy.int8, cupy.uint8]
    funcC(tobytemask, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tobytemask = [False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, True]
    assert cupy.array_equal(tobytemask[:len(pytest_tobytemask)], cupy.array(pytest_tobytemask))




# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_ListArrayU32_getitem_jagged_expand_64_1():
    multistarts = cupy.array([123, 123, 123], dtype=cupy.int64)
    multistops = cupy.array([123, 123, 123], dtype=cupy.int64)
    singleoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromstarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    fromstops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint32)
    jaggedsize = 1
    length = 3
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_expand', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.uint32, cupy.uint32]
    funcC(multistarts, multistops, singleoffsets, tocarry, fromstarts, fromstops, jaggedsize, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_multistarts = [1, 1, 1]
    assert cupy.array_equal(multistarts[:len(pytest_multistarts)], cupy.array(pytest_multistarts))
    pytest_multistops = [1, 1, 1]
    assert cupy.array_equal(multistops[:len(pytest_multistops)], cupy.array(pytest_multistops))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ListArrayU32_getitem_jagged_expand_64_2():
    multistarts = cupy.array([123, 123, 123], dtype=cupy.int64)
    multistops = cupy.array([123, 123, 123], dtype=cupy.int64)
    singleoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromstarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    fromstops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint32)
    jaggedsize = 1
    length = 3
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_expand', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.uint32, cupy.uint32]
    funcC(multistarts, multistops, singleoffsets, tocarry, fromstarts, fromstops, jaggedsize, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_multistarts = [2, 2, 2]
    assert cupy.array_equal(multistarts[:len(pytest_multistarts)], cupy.array(pytest_multistarts))
    pytest_multistops = [3, 3, 3]
    assert cupy.array_equal(multistops[:len(pytest_multistops)], cupy.array(pytest_multistops))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ListArrayU32_getitem_jagged_expand_64_3():
    multistarts = cupy.array([123, 123, 123], dtype=cupy.int64)
    multistops = cupy.array([123, 123, 123], dtype=cupy.int64)
    singleoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromstarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    fromstops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint32)
    jaggedsize = 1
    length = 3
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_expand', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.uint32, cupy.uint32]
    funcC(multistarts, multistops, singleoffsets, tocarry, fromstarts, fromstops, jaggedsize, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_multistarts = [2, 2, 2]
    assert cupy.array_equal(multistarts[:len(pytest_multistarts)], cupy.array(pytest_multistarts))
    pytest_multistops = [1, 1, 1]
    assert cupy.array_equal(multistops[:len(pytest_multistops)], cupy.array(pytest_multistops))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ListArrayU32_getitem_jagged_expand_64_4():
    multistarts = cupy.array([123, 123, 123], dtype=cupy.int64)
    multistops = cupy.array([123, 123, 123], dtype=cupy.int64)
    singleoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromstarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    fromstops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint32)
    jaggedsize = 1
    length = 3
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_expand', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.uint32, cupy.uint32]
    funcC(multistarts, multistops, singleoffsets, tocarry, fromstarts, fromstops, jaggedsize, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_multistarts = [1, 1, 1]
    assert cupy.array_equal(multistarts[:len(pytest_multistarts)], cupy.array(pytest_multistarts))
    pytest_multistops = [0, 0, 0]
    assert cupy.array_equal(multistops[:len(pytest_multistops)], cupy.array(pytest_multistops))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ListArrayU32_getitem_jagged_expand_64_5():
    multistarts = cupy.array([123, 123, 123], dtype=cupy.int64)
    multistops = cupy.array([123, 123, 123], dtype=cupy.int64)
    singleoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromstarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    fromstops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint32)
    jaggedsize = 1
    length = 3
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_expand', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.uint32, cupy.uint32]
    funcC(multistarts, multistops, singleoffsets, tocarry, fromstarts, fromstops, jaggedsize, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_multistarts = [0, 0, 0]
    assert cupy.array_equal(multistarts[:len(pytest_multistarts)], cupy.array(pytest_multistarts))
    pytest_multistops = [0, 0, 0]
    assert cupy.array_equal(multistops[:len(pytest_multistops)], cupy.array(pytest_multistops))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")

string = ak.Array(
    [
        ["\u03b1\u03b2\u03b3", ""],
        [],
        ["\u2192\u03b4\u03b5\u2190", "\u03b6z z\u03b6", "abc"],
    ]
)
bytestring = ak.Array(
    [
        ["\u03b1\u03b2\u03b3".encode(), b""],
        [],
        ["\u2192\u03b4\u03b5\u2190".encode(), "\u03b6z z\u03b6".encode(), b"abc"],
    ]
)

string_padded = ak.Array(
    [
        ["      αβγ      ", "               "],
        [],
        ["     →δε←      ", "     ζz zζ     ", "      abc      "],
    ]
)
bytestring_padded = ak.Array(
    [
        [b"    \xce\xb1\xce\xb2\xce\xb3     ", b"               "],
        [],
        [
            b"  \xe2\x86\x92\xce\xb4\xce\xb5\xe2\x86\x90   ",
            b"    \xce\xb6z z\xce\xb6    ",
            b"      abc      ",
        ],
    ]
)

string_repeats = ak.Array(
    [["foo123bar123baz", "foo", "bar"], ["123foo", "456bar", "foo123456bar"], []]
)

bytestring_repeats = ak.Array(
    [[b"foo123bar123baz", b"foo", b"bar"], [b"123foo", b"456bar", b"foo123456bar"], []]
)


def test_is_alnum():
    assert ak.str.is_alnum(string).tolist() == [
        [True, False],
        [],
        [False, False, True],
    ]
    assert ak.str.is_alnum(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, True],
    ]
    assert (
        ak.str.is_alnum(string).layout.form
        == ak.str.is_alnum(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_alnum(bytestring).layout.form
        == ak.str.is_alnum(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_alpha():
    assert ak.str.is_alpha(string).tolist() == [
        [True, False],
        [],
        [False, False, True],
    ]
    assert ak.str.is_alpha(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, True],
    ]
    assert (
        ak.str.is_alpha(string).layout.form
        == ak.str.is_alpha(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_alpha(bytestring).layout.form
        == ak.str.is_alpha(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_decimal():
    assert ak.str.is_decimal(string).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert ak.str.is_decimal(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert (
        ak.str.is_decimal(string).layout.form
        == ak.str.is_decimal(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_decimal(bytestring).layout.form
        == ak.str.is_decimal(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_digit():
    assert ak.str.is_digit(string).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert ak.str.is_digit(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]

    assert (
        ak.str.is_digit(string).layout.form
        == ak.str.is_digit(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_digit(bytestring).layout.form
        == ak.str.is_digit(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_lower():
    assert ak.str.is_lower(string).tolist() == [
        [True, False],
        [],
        [True, True, True],
    ]
    assert ak.str.is_lower(bytestring).tolist() == [
        [False, False],
        [],
        [False, True, True],
    ]
    assert (
        ak.str.is_lower(string).layout.form
        == ak.str.is_lower(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_lower(bytestring).layout.form
        == ak.str.is_lower(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_numeric():
    assert ak.str.is_numeric(string).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert ak.str.is_numeric(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert (
        ak.str.is_numeric(string).layout.form
        == ak.str.is_numeric(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_numeric(bytestring).layout.form
        == ak.str.is_numeric(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_printable():
    assert ak.str.is_printable(string).tolist() == [
        [True, True],
        [],
        [True, True, True],
    ]
    assert ak.str.is_printable(bytestring).tolist() == [
        [False, True],
        [],
        [False, False, True],
    ]
    assert (
        ak.str.is_printable(string).layout.form
        == ak.str.is_printable(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_printable(bytestring).layout.form
        == ak.str.is_printable(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_space():
    assert ak.str.is_space(string).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert ak.str.is_space(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert (
        ak.str.is_space(string).layout.form
        == ak.str.is_space(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_space(bytestring).layout.form
        == ak.str.is_space(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_upper():
    assert ak.str.is_upper(string).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert ak.str.is_upper(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert (
        ak.str.is_upper(string).layout.form
        == ak.str.is_upper(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_upper(bytestring).layout.form
        == ak.str.is_upper(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_title():
    assert ak.str.is_title(string).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert ak.str.is_title(bytestring).tolist() == [
        [False, False],
        [],
        [False, False, False],
    ]
    assert (
        ak.str.is_title(string).layout.form
        == ak.str.is_title(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_title(bytestring).layout.form
        == ak.str.is_title(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_is_ascii():
    assert ak.str.is_ascii(string).tolist() == [
        [False, True],
        [],
        [False, False, True],
    ]
    assert ak.str.is_ascii(bytestring).tolist() == [
        [False, True],
        [],
        [False, False, True],
    ]
    assert (
        ak.str.is_ascii(string).layout.form
        == ak.str.is_ascii(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.is_ascii(bytestring).layout.form
        == ak.str.is_ascii(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_capitalize():
    assert ak.str.capitalize(string).tolist() == [
        ["Αβγ", ""],
        [],
        ["→δε←", "Ζz zζ", "Abc"],  # noqa: RUF001, RUF003 (we care about Ζ vs Z)
    ]
    assert ak.str.capitalize(bytestring).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"Abc"],
    ]
    assert (
        ak.str.capitalize(string).layout.form
        == ak.str.capitalize(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.capitalize(bytestring).layout.form
        == ak.str.capitalize(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_length():
    assert ak.str.length(string).tolist() == [
        [3, 0],
        [],
        [4, 5, 3],
    ]
    assert ak.str.length(bytestring).tolist() == [
        [6, 0],
        [],
        [10, 7, 3],
    ]
    assert (
        ak.str.length(string).layout.form
        == ak.str.length(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.length(bytestring).layout.form
        == ak.str.length(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_lower():
    assert ak.str.lower(string).tolist() == [
        ["αβγ", ""],
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.lower(bytestring).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.lower(string).layout.form
        == ak.str.lower(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.lower(bytestring).layout.form
        == ak.str.lower(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_swapcase():
    assert ak.str.swapcase(string).tolist() == [
        ["ΑΒΓ", ""],
        [],
        ["→ΔΕ←", "ΖZ ZΖ", "ABC"],  # noqa: RUF001, RUF003 (we care about Ζ vs Z)
    ]
    assert ak.str.swapcase(bytestring).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζZ Zζ".encode(), b"ABC"],
    ]
    assert (
        ak.str.swapcase(string).layout.form
        == ak.str.swapcase(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.swapcase(bytestring).layout.form
        == ak.str.swapcase(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_title():
    assert ak.str.title(string).tolist() == [
        ["Αβγ", ""],
        [],
        ["→Δε←", "Ζz Zζ", "Abc"],  # noqa: RUF001, RUF003 (we care about Ζ vs Z)
    ]
    assert ak.str.title(bytestring).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζZ Zζ".encode(), b"Abc"],
    ]
    assert (
        ak.str.title(string).layout.form
        == ak.str.title(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.title(bytestring).layout.form
        == ak.str.title(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_upper():
    assert ak.str.upper(string).tolist() == [
        ["ΑΒΓ", ""],
        [],
        ["→ΔΕ←", "ΖZ ZΖ", "ABC"],  # noqa: RUF001, RUF003 (we care about Ζ vs Z)
    ]
    assert ak.str.upper(bytestring).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζZ Zζ".encode(), b"ABC"],
    ]
    assert (
        ak.str.upper(string).layout.form
        == ak.str.upper(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.upper(bytestring).layout.form
        == ak.str.upper(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_repeat():
    assert ak.str.repeat(string, 3).tolist() == [
        ["αβγαβγαβγ", ""],
        [],
        ["→δε←→δε←→δε←", "ζz zζζz zζζz zζ", "abcabcabc"],
    ]
    assert ak.str.repeat(bytestring, 3).tolist() == [
        ["αβγαβγαβγ".encode(), b""],
        [],
        ["→δε←→δε←→δε←".encode(), "ζz zζζz zζζz zζ".encode(), b"abcabcabc"],
    ]
    assert (
        ak.str.repeat(string, 3).layout.form
        == ak.str.repeat(ak.to_backend(string, "typetracer"), 3).layout.form
    )
    assert (
        ak.str.repeat(bytestring, 3).layout.form
        == ak.str.repeat(ak.to_backend(bytestring, "typetracer"), 3).layout.form
    )

    assert ak.str.repeat(string, [[3, 3], [], [2, 0, 1]]).tolist() == [
        ["αβγαβγαβγ", ""],
        [],
        ["→δε←→δε←", "", "abc"],
    ]
    assert ak.str.repeat(bytestring, [[3, 3], [], [2, 0, 1]]).tolist() == [
        ["αβγαβγαβγ".encode(), b""],
        [],
        ["→δε←→δε←".encode(), b"", b"abc"],
    ]
    assert (
        ak.str.repeat(string, [[3, 3], [], [2, 0, 1]]).layout.form
        == ak.str.repeat(
            ak.to_backend(string, "typetracer"), [[3, 3], [], [2, 0, 1]]
        ).layout.form
    )
    assert (
        ak.str.repeat(bytestring, [[3, 3], [], [2, 0, 1]]).layout.form
        == ak.str.repeat(
            ak.to_backend(bytestring, "typetracer"), [[3, 3], [], [2, 0, 1]]
        ).layout.form
    )


def test_replace_slice():
    assert ak.str.replace_slice(string[:, :1], 1, 2, "qj").tolist() == [
        ["αqjγ"],  # noqa: RUF001
        [],
        ["→qjε←"],
    ]
    assert ak.str.replace_slice(bytestring[:, :1], 1, 2, b"qj").tolist() == [
        [b"\xceqj\xce\xb2\xce\xb3"],
        [],
        [b"\xe2qj\x92\xce\xb4\xce\xb5\xe2\x86\x90"],
    ]
    assert (
        ak.str.replace_slice(string[:, :1], 1, 2, "qj").layout.form
        == ak.str.replace_slice(
            ak.to_backend(string, "typetracer")[:, :1], 1, 2, "qj"
        ).layout.form
    )
    assert (
        ak.str.replace_slice(bytestring[:, :1], 1, 2, b"qj").layout.form
        == ak.str.replace_slice(
            ak.to_backend(bytestring, "typetracer")[:, :1], 1, 2, b"qj"
        ).layout.form
    )


def test_reverse():
    assert ak.str.reverse(string).tolist() == [
        ["αβγ"[::-1], ""],
        [],
        ["→δε←"[::-1], "ζz zζ"[::-1], "abc"[::-1]],
    ]
    assert ak.str.reverse(bytestring).tolist() == [
        ["αβγ".encode()[::-1], b""],
        [],
        ["→δε←".encode()[::-1], "ζz zζ".encode()[::-1], b"abc"[::-1]],
    ]
    assert (
        ak.str.reverse(string).layout.form
        == ak.str.reverse(ak.to_backend(string, "typetracer")).layout.form
    )
    assert (
        ak.str.reverse(bytestring).layout.form
        == ak.str.reverse(ak.to_backend(bytestring, "typetracer")).layout.form
    )


def test_replace_substring():
    assert ak.str.replace_substring(string, "βγ", "HELLO").tolist() == [
        ["αHELLO", ""],  # noqa: RUF001
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.replace_substring(bytestring, "βγ".encode(), b"HELLO").tolist() == [
        ["αHELLO".encode(), b""],  # noqa: RUF001
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.replace_substring(string, "βγ", "HELLO").layout.form
        == ak.str.replace_substring(
            ak.to_backend(string, "typetracer"), "βγ", "HELLO"
        ).layout.form
    )
    assert (
        ak.str.replace_substring(bytestring, "βγ".encode(), b"HELLO").layout.form
        == ak.str.replace_substring(
            ak.to_backend(bytestring, "typetracer"), "βγ".encode(), b"HELLO"
        ).layout.form
    )

    assert ak.str.replace_substring(
        string, "βγ", "HELLO", max_replacements=0
    ).tolist() == [
        ["αβγ", ""],
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.replace_substring(
        bytestring, "βγ".encode(), b"HELLO", max_replacements=0
    ).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.replace_substring(string, "βγ", "HELLO", max_replacements=0).layout.form
        == ak.str.replace_substring(
            ak.to_backend(string, "typetracer"), "βγ", "HELLO", max_replacements=0
        ).layout.form
    )
    assert (
        ak.str.replace_substring(
            bytestring, "βγ".encode(), b"HELLO", max_replacements=0
        ).layout.form
        == ak.str.replace_substring(
            ak.to_backend(bytestring, "typetracer"),
            "βγ".encode(),
            b"HELLO",
            max_replacements=0,
        ).layout.form
    )


def test_replace_substring_regex():
    assert ak.str.replace_substring_regex(string, "βγ", "HELLO").tolist() == [
        ["αHELLO", ""],  # noqa: RUF001
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.replace_substring_regex(
        bytestring, "βγ".encode(), b"HELLO"
    ).tolist() == [
        ["αHELLO".encode(), b""],  # noqa: RUF001
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.replace_substring(string, "βγ", "HELLO").layout.form
        == ak.str.replace_substring(
            ak.to_backend(string, "typetracer"), "βγ", "HELLO"
        ).layout.form
    )
    assert (
        ak.str.replace_substring(bytestring, "βγ".encode(), b"HELLO").layout.form
        == ak.str.replace_substring(
            ak.to_backend(bytestring, "typetracer"), "βγ".encode(), b"HELLO"
        ).layout.form
    )

    assert ak.str.replace_substring_regex(
        string, "βγ", "HELLO", max_replacements=0
    ).tolist() == [
        ["αβγ", ""],
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.replace_substring_regex(
        bytestring, "βγ".encode(), b"HELLO", max_replacements=0
    ).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.replace_substring(string, "βγ", "HELLO", max_replacements=0).layout.form
        == ak.str.replace_substring(
            ak.to_backend(string, "typetracer"), "βγ", "HELLO", max_replacements=0
        ).layout.form
    )
    assert (
        ak.str.replace_substring(
            bytestring, "βγ".encode(), b"HELLO", max_replacements=0
        ).layout.form
        == ak.str.replace_substring(
            ak.to_backend(bytestring, "typetracer"),
            "βγ".encode(),
            b"HELLO",
            max_replacements=0,
        ).layout.form
    )

    # Check regex
    assert ak.str.replace_substring_regex(
        [["aaaa1bb1c2ddddd"], ["fff"], []], r"[a-zA-Z]+", "FOO"
    ).tolist() == [["FOO1FOO1FOO2FOO"], ["FOO"], []]
    assert ak.str.replace_substring_regex(
        [[b"aaaa1bb1c2ddddd"], [b"fff"], []], rb"[a-zA-Z]+", b"FOO"
    ).tolist() == [[b"FOO1FOO1FOO2FOO"], [b"FOO"], []]
    assert (
        ak.str.replace_substring_regex(
            [["aaaa1bb1c2ddddd"], ["fff"], []], r"[a-zA-Z]+", "FOO"
        ).layout.form
        == ak.str.replace_substring_regex(
            ak.to_backend([["aaaa1bb1c2ddddd"], ["fff"], []], "typetracer"),
            r"[a-zA-Z]+",
            "FOO",
        ).layout.form
    )
    assert (
        ak.str.replace_substring_regex(
            [[b"aaaa1bb1c2ddddd"], [b"fff"], []], rb"[a-zA-Z]+", b"FOO"
        ).layout.form
        == ak.str.replace_substring_regex(
            ak.to_backend([[b"aaaa1bb1c2ddddd"], [b"fff"], []], "typetracer"),
            rb"[a-zA-Z]+",
            b"FOO",
        ).layout.form
    )


def test_center():
    assert ak.str.center(string, 15, " ").tolist() == [
        ["      αβγ      ", "               "],
        [],
        ["     →δε←      ", "     ζz zζ     ", "      abc      "],
    ]
    assert ak.str.center(bytestring, 15, b" ").tolist() == [
        [b"    \xce\xb1\xce\xb2\xce\xb3     ", b"               "],
        [],
        [
            b"  \xe2\x86\x92\xce\xb4\xce\xb5\xe2\x86\x90   ",
            b"    \xce\xb6z z\xce\xb6    ",
            b"      abc      ",
        ],
    ]
    assert (
        ak.str.center(string, 15, " ").layout.form
        == ak.str.center(ak.to_backend(string, "typetracer"), 15, " ").layout.form
    )
    assert (
        ak.str.center(bytestring, 15, b" ").layout.form
        == ak.str.center(ak.to_backend(bytestring, "typetracer"), 15, b" ").layout.form
    )


def test_lpad():
    assert ak.str.lpad(string, 15, " ").tolist() == [
        ["            αβγ", "               "],
        [],
        ["           →δε←", "          ζz zζ", "            abc"],
    ]
    assert ak.str.lpad(bytestring, 15, b" ").tolist() == [
        [b"         \xce\xb1\xce\xb2\xce\xb3", b"               "],
        [],
        [
            b"     \xe2\x86\x92\xce\xb4\xce\xb5\xe2\x86\x90",
            b"        \xce\xb6z z\xce\xb6",
            b"            abc",
        ],
    ]
    assert (
        ak.str.lpad(string, 15, " ").layout.form
        == ak.str.lpad(ak.to_backend(string, "typetracer"), 15, " ").layout.form
    )
    assert (
        ak.str.lpad(bytestring, 15, b" ").layout.form
        == ak.str.lpad(ak.to_backend(bytestring, "typetracer"), 15, b" ").layout.form
    )


def test_rpad():
    assert ak.str.rpad(string, 15, " ").tolist() == [
        ["αβγ            ", "               "],
        [],
        ["→δε←           ", "ζz zζ          ", "abc            "],
    ]
    assert ak.str.rpad(bytestring, 15, b" ").tolist() == [
        [b"\xce\xb1\xce\xb2\xce\xb3         ", b"               "],
        [],
        [
            b"\xe2\x86\x92\xce\xb4\xce\xb5\xe2\x86\x90     ",
            b"\xce\xb6z z\xce\xb6        ",
            b"abc            ",
        ],
    ]
    assert (
        ak.str.rpad(string, 15, " ").layout.form
        == ak.str.rpad(ak.to_backend(string, "typetracer"), 15, " ").layout.form
    )
    assert (
        ak.str.rpad(bytestring, 15, b" ").layout.form
        == ak.str.rpad(ak.to_backend(bytestring, "typetracer"), 15, b" ").layout.form
    )


def test_ltrim():
    assert ak.str.ltrim(string_padded, " ").tolist() == [
        ["αβγ      ", ""],
        [],
        ["→δε←      ", "ζz zζ     ", "abc      "],
    ]
    assert ak.str.ltrim(bytestring_padded, b" ").tolist() == [
        ["αβγ     ".encode(), b""],
        [],
        ["→δε←   ".encode(), "ζz zζ    ".encode(), b"abc      "],
    ]
    assert (
        ak.str.ltrim(string_padded, " ").layout.form
        == ak.str.ltrim(ak.to_backend(string_padded, "typetracer"), " ").layout.form
    )
    assert (
        ak.str.ltrim(bytestring_padded, b" ").layout.form
        == ak.str.ltrim(
            ak.to_backend(bytestring_padded, "typetracer"), b" "
        ).layout.form
    )


def test_ltrim_whitespace():
    assert ak.str.ltrim_whitespace(string_padded).tolist() == [
        ["αβγ      ", ""],
        [],
        ["→δε←      ", "ζz zζ     ", "abc      "],
    ]
    assert ak.str.ltrim_whitespace(bytestring_padded).tolist() == [
        ["αβγ     ".encode(), b""],
        [],
        ["→δε←   ".encode(), "ζz zζ    ".encode(), b"abc      "],
    ]
    assert (
        ak.str.ltrim_whitespace(string_padded).layout.form
        == ak.str.ltrim_whitespace(
            ak.to_backend(string_padded, "typetracer")
        ).layout.form
    )
    assert (
        ak.str.ltrim_whitespace(bytestring_padded).layout.form
        == ak.str.ltrim_whitespace(
            ak.to_backend(bytestring_padded, "typetracer")
        ).layout.form
    )


def test_rtrim():
    assert ak.str.rtrim(string_padded, " ").tolist() == [
        ["      αβγ", ""],
        [],
        ["     →δε←", "     ζz zζ", "      abc"],
    ]
    assert ak.str.rtrim(bytestring_padded, b" ").tolist() == [
        ["    αβγ".encode(), b""],
        [],
        ["  →δε←".encode(), "    ζz zζ".encode(), b"      abc"],
    ]
    assert (
        ak.str.rtrim(string_padded, " ").layout.form
        == ak.str.rtrim(ak.to_backend(string_padded, "typetracer"), " ").layout.form
    )
    assert (
        ak.str.rtrim(bytestring_padded, b" ").layout.form
        == ak.str.rtrim(
            ak.to_backend(bytestring_padded, "typetracer"), b" "
        ).layout.form
    )


def test_rtrim_whitespace():
    assert ak.str.rtrim_whitespace(string_padded).tolist() == [
        ["      αβγ", ""],
        [],
        ["     →δε←", "     ζz zζ", "      abc"],
    ]
    assert ak.str.rtrim_whitespace(bytestring_padded).tolist() == [
        ["    αβγ".encode(), b""],
        [],
        ["  →δε←".encode(), "    ζz zζ".encode(), b"      abc"],
    ]
    assert (
        ak.str.rtrim_whitespace(string_padded).layout.form
        == ak.str.rtrim_whitespace(
            ak.to_backend(string_padded, "typetracer")
        ).layout.form
    )
    assert (
        ak.str.rtrim_whitespace(bytestring_padded).layout.form
        == ak.str.rtrim_whitespace(
            ak.to_backend(bytestring_padded, "typetracer")
        ).layout.form
    )


def test_trim():
    assert ak.str.trim(string_padded, " ").tolist() == [
        ["αβγ", ""],
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.trim(bytestring_padded, b" ").tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.trim(string_padded, " ").layout.form
        == ak.str.trim(ak.to_backend(string_padded, "typetracer"), " ").layout.form
    )
    assert (
        ak.str.trim(bytestring_padded, b" ").layout.form
        == ak.str.trim(ak.to_backend(bytestring_padded, "typetracer"), b" ").layout.form
    )


def test_trim_whitespace():
    assert ak.str.trim_whitespace(string_padded).tolist() == [
        ["αβγ", ""],
        [],
        ["→δε←", "ζz zζ", "abc"],
    ]
    assert ak.str.trim_whitespace(bytestring_padded).tolist() == [
        ["αβγ".encode(), b""],
        [],
        ["→δε←".encode(), "ζz zζ".encode(), b"abc"],
    ]
    assert (
        ak.str.trim_whitespace(string_padded).layout.form
        == ak.str.trim_whitespace(
            ak.to_backend(string_padded, "typetracer")
        ).layout.form
    )

    assert (
        ak.str.trim_whitespace(bytestring_padded).layout.form
        == ak.str.trim_whitespace(
            ak.to_backend(bytestring_padded, "typetracer")
        ).layout.form
    )


def test_slice():
    assert ak.str.slice(string, 1, 3).tolist() == [
        ["αβγ"[1:3], ""[1:3]],
        [],
        ["→δε←"[1:3], "ζz zζ"[1:3], "abc"[1:3]],
    ]
    assert ak.str.slice(bytestring, 1, 3).tolist() == [
        ["αβγ".encode()[1:3], b""[1:3]],
        [],
        ["→δε←".encode()[1:3], "ζz zζ".encode()[1:3], b"abc"[1:3]],
    ]
    assert (
        ak.str.slice(string, 1, 3).layout.form
        == ak.str.slice(ak.to_backend(string, "typetracer"), 1, 3).layout.form
    )
    assert (
        ak.str.slice(bytestring, 1, 3).layout.form
        == ak.str.slice(ak.to_backend(bytestring, "typetracer"), 1, 3).layout.form
    )

    # ArrowInvalid: Negative buffer resize: -40 (looks like an Arrow bug)
    # assert ak.str.slice(string, 1).tolist() == [
    #     ["αβγ"[1:], ""[1:]],
    #     [],
    #     ["→δε←"[1:], "ζz zζ"[1:], "abc"[1:]],
    # ]
    # assert (
    #     ak.str.slice(string, 1).layout.form
    #     == ak.str.slice(ak.to_backend(string, "typetracer"), 1).layout.form
    # )
    assert ak.str.slice(bytestring, 1).tolist() == [
        ["αβγ".encode()[1:], b""[1:]],
        [],
        ["→δε←".encode()[1:], "ζz zζ".encode()[1:], b"abc"[1:]],
    ]
    assert (
        ak.str.slice(bytestring, 1).layout.form
        == ak.str.slice(ak.to_backend(bytestring, "typetracer"), 1).layout.form
    )


def test_split_whitespace():
    assert ak.str.split_whitespace(string_padded, max_splits=1).tolist() == [
        [["", "αβγ      "], ["", " "]],
        [],
        [["", "→δε←      "], ["", "ζz zζ     "], ["", "abc      "]],
    ]
    assert (
        ak.str.split_whitespace(string_padded, max_splits=1).layout.form
        == ak.str.split_whitespace(
            ak.to_backend(string_padded, "typetracer"), max_splits=1
        ).layout.form
    )

    assert ak.str.split_whitespace(
        string_padded, max_splits=1, reverse=True
    ).tolist() == [
        [["      αβγ", ""], [" ", ""]],
        [],
        [["     →δε←", ""], ["     ζz zζ", ""], ["      abc", ""]],
    ]
    assert (
        ak.str.split_whitespace(string_padded, max_splits=1, reverse=True).layout.form
        == ak.str.split_whitespace(
            ak.to_backend(string_padded, "typetracer"), max_splits=1, reverse=True
        ).layout.form
    )

    assert ak.str.split_whitespace(string_padded, max_splits=None).tolist() == [
        [["", "αβγ", "", ""], ["", "", ""]],
        [],
        [["", "→δε←", "", ""], ["", "ζz", "zζ", "", ""], ["", "abc", "", ""]],
    ]
    assert (
        ak.str.split_whitespace(string_padded, max_splits=None).layout.form
        == ak.str.split_whitespace(
            ak.to_backend(string_padded, "typetracer"), max_splits=None
        ).layout.form
    )

    # Bytestrings
    assert ak.str.split_whitespace(bytestring_padded, max_splits=1).tolist() == [
        [[b"", "αβγ     ".encode()], [b"", b""]],
        [],
        [
            [b"", "→δε←   ".encode()],
            [b"", "ζz zζ    ".encode()],
            [b"", b"abc      "],
        ],
    ]
    assert (
        ak.str.split_whitespace(bytestring_padded, max_splits=1).layout.form
        == ak.str.split_whitespace(
            ak.to_backend(bytestring_padded, "typetracer"), max_splits=1
        ).layout.form
    )

    assert ak.str.split_whitespace(
        bytestring_padded, max_splits=1, reverse=True
    ).tolist() == [
        [["    αβγ".encode(), b""], [b"", b""]],
        [],
        [
            ["  →δε←".encode(), b""],
            ["    ζz zζ".encode(), b""],
            [b"      abc", b""],
        ],
    ]
    assert (
        ak.str.split_whitespace(
            bytestring_padded, max_splits=1, reverse=True
        ).layout.form
        == ak.str.split_whitespace(
            ak.to_backend(bytestring_padded, "typetracer"), max_splits=1, reverse=True
        ).layout.form
    )

    assert ak.str.split_whitespace(bytestring_padded, max_splits=None).tolist() == [
        [[b"", "αβγ".encode(), b""], [b"", b""]],
        [],
        [
            [b"", "→δε←".encode(), b""],
            [b"", "ζz".encode(), "zζ".encode(), b""],
            [b"", b"abc", b""],
        ],
    ]
    assert (
        ak.str.split_whitespace(bytestring_padded, max_splits=None).layout.form
        == ak.str.split_whitespace(
            ak.to_backend(bytestring_padded, "typetracer"), max_splits=None
        ).layout.form
    )


def test_split_pattern():
    assert ak.str.split_pattern(string_repeats, "123", max_splits=1).tolist() == [
        [["foo", "bar123baz"], ["foo"], ["bar"]],
        [["", "foo"], ["456bar"], ["foo", "456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern(string_repeats, "123", max_splits=1).layout.form
        == ak.str.split_pattern(
            ak.to_backend(string_repeats, "typetracer"), "123", max_splits=1
        ).layout.form
    )

    assert ak.str.split_pattern(
        string_repeats, "123", max_splits=1, reverse=True
    ).tolist() == [
        [["foo123bar", "baz"], ["foo"], ["bar"]],
        [["", "foo"], ["456bar"], ["foo", "456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern(
            string_repeats, "123", max_splits=1, reverse=True
        ).layout.form
        == ak.str.split_pattern(
            ak.to_backend(string_repeats, "typetracer"),
            "123",
            max_splits=1,
            reverse=True,
        ).layout.form
    )
    assert ak.str.split_pattern(string_repeats, "123", max_splits=None).tolist() == [
        [["foo", "bar", "baz"], ["foo"], ["bar"]],
        [["", "foo"], ["456bar"], ["foo", "456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern(string_repeats, "123", max_splits=None).layout.form
        == ak.str.split_pattern(
            ak.to_backend(string_repeats, "typetracer"), "123", max_splits=None
        ).layout.form
    )

    # Bytestrings
    assert ak.str.split_pattern(bytestring_repeats, b"123", max_splits=1).tolist() == [
        [[b"foo", b"bar123baz"], [b"foo"], [b"bar"]],
        [[b"", b"foo"], [b"456bar"], [b"foo", b"456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern(bytestring_repeats, b"123", max_splits=1).layout.form
        == ak.str.split_pattern(
            ak.to_backend(bytestring_repeats, "typetracer"), b"123", max_splits=1
        ).layout.form
    )

    assert ak.str.split_pattern(
        bytestring_repeats, b"123", max_splits=1, reverse=True
    ).tolist() == [
        [[b"foo123bar", b"baz"], [b"foo"], [b"bar"]],
        [[b"", b"foo"], [b"456bar"], [b"foo", b"456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern(
            bytestring_repeats, b"123", max_splits=1, reverse=True
        ).layout.form
        == ak.str.split_pattern(
            ak.to_backend(bytestring_repeats, "typetracer"),
            b"123",
            max_splits=1,
            reverse=True,
        ).layout.form
    )

    assert ak.str.split_pattern(
        bytestring_repeats, b"123", max_splits=None
    ).tolist() == [
        [[b"foo", b"bar", b"baz"], [b"foo"], [b"bar"]],
        [[b"", b"foo"], [b"456bar"], [b"foo", b"456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern(bytestring_repeats, b"123", max_splits=None).layout.form
        == ak.str.split_pattern(
            ak.to_backend(bytestring_repeats, "typetracer"), b"123", max_splits=None
        ).layout.form
    )


def test_split_pattern_regex():
    assert ak.str.split_pattern_regex(
        string_repeats, r"\d{3}", max_splits=1
    ).tolist() == [
        [["foo", "bar123baz"], ["foo"], ["bar"]],
        [["", "foo"], ["", "bar"], ["foo", "456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern_regex(string_repeats, r"\d{3}", max_splits=1).layout.form
        == ak.str.split_pattern_regex(
            ak.to_backend(string_repeats, "typetracer"), r"\d{3}", max_splits=1
        ).layout.form
    )

    with pytest.raises(ValueError, match=r"split in reverse with regex"):
        assert ak.str.split_pattern_regex(
            string_repeats, r"\d{3}", max_splits=1, reverse=True
        ).tolist() == [
            [["foo123bar", "baz"], ["foo"], ["bar"]],
            [["", "foo"], ["", "bar"], ["foo", "456bar"]],
            [],
        ]
    with pytest.raises(ValueError, match=r"split in reverse with regex"):
        ak.str.split_pattern_regex(
            ak.to_backend(string_repeats, "typetracer"),
            r"\d{3}",
            max_splits=1,
            reverse=True,
        )

    assert ak.str.split_pattern_regex(
        string_repeats, r"\d{3}", max_splits=None
    ).tolist() == [
        [["foo", "bar", "baz"], ["foo"], ["bar"]],
        [["", "foo"], ["", "bar"], ["foo", "", "bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern_regex(
            string_repeats, r"\d{3}", max_splits=None
        ).layout.form
        == ak.str.split_pattern_regex(
            ak.to_backend(string_repeats, "typetracer"), r"\d{3}", max_splits=None
        ).layout.form
    )

    # Bytestrings
    assert ak.str.split_pattern_regex(
        bytestring_repeats, rb"\d{3}", max_splits=1
    ).tolist() == [
        [[b"foo", b"bar123baz"], [b"foo"], [b"bar"]],
        [[b"", b"foo"], [b"", b"bar"], [b"foo", b"456bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern_regex(
            bytestring_repeats, r"\d{3}", max_splits=1
        ).layout.form
        == ak.str.split_pattern_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), r"\d{3}", max_splits=1
        ).layout.form
    )

    with pytest.raises(ValueError, match=r"split in reverse with regex"):
        assert ak.str.split_pattern_regex(
            bytestring_repeats, rb"\d{3}", max_splits=1, reverse=True
        ).tolist() == [
            [[b"foo123bar", b"baz"], [b"foo"], [b"bar"]],
            [[b"", b"foo"], [b"", b"bar"], [b"foo", b"456bar"]],
            [],
        ]
    with pytest.raises(ValueError, match=r"split in reverse with regex"):
        ak.str.split_pattern_regex(
            ak.to_backend(bytestring_repeats, "typetracer"),
            r"\d{3}",
            max_splits=1,
            reverse=True,
        )

    assert ak.str.split_pattern_regex(
        bytestring_repeats, rb"\d{3}", max_splits=None
    ).tolist() == [
        [[b"foo", b"bar", b"baz"], [b"foo"], [b"bar"]],
        [[b"", b"foo"], [b"", b"bar"], [b"foo", b"", b"bar"]],
        [],
    ]
    assert (
        ak.str.split_pattern_regex(
            bytestring_repeats, r"\d{3}", max_splits=None
        ).layout.form
        == ak.str.split_pattern_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), r"\d{3}", max_splits=None
        ).layout.form
    )


def test_extract_regex():
    assert ak.str.extract_regex(
        ak.Array([["one1", "two2", "three3"], [], ["four4", "five5"]]),
        "(?P<vowel>[aeiou])(?P<number>[0-9]+)",
    ).tolist() == [
        [
            {"vowel": "e", "number": "1"},
            {"vowel": "o", "number": "2"},
            {"vowel": "e", "number": "3"},
        ],
        [],
        [None, {"vowel": "e", "number": "5"}],
    ]
    assert (
        ak.str.extract_regex(
            ak.Array([["one1", "two2", "three3"], [], ["four4", "five5"]]),
            "(?P<vowel>[aeiou])(?P<number>[0-9]+)",
        ).layout.form
        == ak.str.extract_regex(
            ak.Array(
                [["one1", "two2", "three3"], [], ["four4", "five5"]],
                backend="typetracer",
            ),
            "(?P<vowel>[aeiou])(?P<number>[0-9]+)",
        ).layout.form
    )

    assert ak.str.extract_regex(
        ak.Array([[b"one1", b"two2", b"three3"], [], [b"four4", b"five5"]]),
        b"(?P<vowel>[aeiou])(?P<number>[0-9]+)",
    ).tolist() == [
        [
            {"vowel": b"e", "number": b"1"},
            {"vowel": b"o", "number": b"2"},
            {"vowel": b"e", "number": b"3"},
        ],
        [],
        [None, {"vowel": b"e", "number": b"5"}],
    ]
    assert (
        ak.str.extract_regex(
            ak.Array([[b"one1", b"two2", b"three3"], [], [b"four4", b"five5"]]),
            b"(?P<vowel>[aeiou])(?P<number>[0-9]+)",
        ).layout.form
        == ak.str.extract_regex(
            ak.Array(
                [[b"one1", b"two2", b"three3"], [], [b"four4", b"five5"]],
                backend="typetracer",
            ),
            b"(?P<vowel>[aeiou])(?P<number>[0-9]+)",
        ).layout.form
    )


def test_join():
    array1 = ak.Array(
        [
            ["this", "that"],
            [],
            ["foo", "bar", "baz"],
        ]
    )
    assert ak.str.join(array1, "-").tolist() == ["this-that", "", "foo-bar-baz"]
    assert (
        ak.str.join(array1, "-").layout.form
        == ak.str.join(ak.to_backend(array1, "typetracer"), "-").layout.form
    )

    separator = ak.Array(["→", "↑", "←"])
    assert ak.str.join(array1, separator).tolist() == ["this→that", "", "foo←bar←baz"]
    assert (
        ak.str.join(array1, separator).layout.form
        == ak.str.join(ak.to_backend(array1, "typetracer"), separator).layout.form
    )

    array2 = ak.Array(
        [
            [b"this", b"that"],
            [],
            [b"foo", b"bar", b"baz"],
        ]
    )
    assert ak.str.join(array2, b"-").tolist() == [b"this-that", b"", b"foo-bar-baz"]
    assert (
        ak.str.join(array2, b"-").layout.form
        == ak.str.join(ak.to_backend(array2, "typetracer"), b"-").layout.form
    )

    separator2 = ak.Array(["→".encode(), "↑".encode(), "←".encode()])
    assert ak.str.join(array2, separator2).tolist() == [
        "this→that".encode(),
        b"",
        "foo←bar←baz".encode(),
    ]
    assert (
        ak.str.join(array2, separator2).layout.form
        == ak.str.join(ak.to_backend(array2, "typetracer"), separator2).layout.form
    )


def test_join_element_wise():
    first1 = ak.Array([["one", "two", "three"], [], ["four", "five"]])
    second1 = ak.Array([["111", "222", "333"], [], ["444", "555"]])
    separator1 = ak.Array(["→", "↑", "←"])

    assert ak.str.join_element_wise(first1, second1, separator1).tolist() == [
        ["one→111", "two→222", "three→333"],
        [],
        ["four←444", "five←555"],
    ]
    assert (
        ak.str.join_element_wise(first1, second1, separator1).layout.form
        == ak.str.join_element_wise(
            ak.to_backend(first1, "typetracer"),
            ak.to_backend(second1, "typetracer"),
            ak.to_backend(separator1, "typetracer"),
        ).layout.form
    )

    first2 = ak.Array([[b"one", b"two", b"three"], [], [b"four", b"five"]])
    second2 = ak.Array([[b"111", b"222", b"333"], [], [b"444", b"555"]])
    separator2 = ak.Array(["→".encode(), "↑".encode(), "←".encode()])

    assert ak.str.join_element_wise(first2, second2, separator2).tolist() == [
        ["one→111".encode(), "two→222".encode(), "three→333".encode()],
        [],
        ["four←444".encode(), "five←555".encode()],
    ]
    assert (
        ak.str.join_element_wise(first2, second2, separator2).layout.form
        == ak.str.join_element_wise(
            ak.to_backend(first2, "typetracer"),
            ak.to_backend(second2, "typetracer"),
            ak.to_backend(separator2, "typetracer"),
        ).layout.form
    )


def test_count_substring():
    assert ak.str.count_substring(string_repeats, "BA").tolist() == [
        [0, 0, 0],
        [0, 0, 0],
        [],
    ]
    assert (
        ak.str.count_substring(string_repeats, "BA").layout.form
        == ak.str.count_substring(
            ak.to_backend(string_repeats, "typetracer"), "BA"
        ).layout.form
    )

    assert ak.str.count_substring(string_repeats, "BA", ignore_case=True).tolist() == [
        [2, 0, 1],
        [0, 1, 1],
        [],
    ]
    assert (
        ak.str.count_substring(string_repeats, "BA", ignore_case=True).layout.form
        == ak.str.count_substring(
            ak.to_backend(string_repeats, "typetracer"), "BA", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.count_substring(bytestring_repeats, b"BA").tolist() == [
        [0, 0, 0],
        [0, 0, 0],
        [],
    ]
    assert (
        ak.str.count_substring(bytestring_repeats, b"BA").layout.form
        == ak.str.count_substring(
            ak.to_backend(bytestring_repeats, "typetracer"), b"BA"
        ).layout.form
    )

    assert ak.str.count_substring(
        bytestring_repeats, b"BA", ignore_case=True
    ).tolist() == [
        [2, 0, 1],
        [0, 1, 1],
        [],
    ]
    assert (
        ak.str.count_substring(bytestring_repeats, b"BA", ignore_case=True).layout.form
        == ak.str.count_substring(
            ak.to_backend(bytestring_repeats, "typetracer"), b"BA", ignore_case=True
        ).layout.form
    )


def test_count_substring_regex():
    assert ak.str.count_substring_regex(string_repeats, r"BA\d*").tolist() == [
        [0, 0, 0],
        [0, 0, 0],
        [],
    ]
    assert (
        ak.str.count_substring_regex(string_repeats, r"BA\d*").layout.form
        == ak.str.count_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"BA\d*"
        ).layout.form
    )

    assert ak.str.count_substring_regex(
        string_repeats, r"BA\d*", ignore_case=True
    ).tolist() == [
        [2, 0, 1],
        [0, 1, 1],
        [],
    ]
    assert (
        ak.str.count_substring_regex(
            string_repeats, r"BA\d*", ignore_case=True
        ).layout.form
        == ak.str.count_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"BA\d*", ignore_case=True
        ).layout.form
    )

    assert ak.str.count_substring_regex(string_repeats, r"\d{1,}").tolist() == [
        [2, 0, 0],
        [1, 1, 1],
        [],
    ]
    assert (
        ak.str.count_substring_regex(string_repeats, r"\d{1,}").layout.form
        == ak.str.count_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"\d{1,}"
        ).layout.form
    )

    # Bytestrings
    assert ak.str.count_substring_regex(bytestring_repeats, rb"BA\d*").tolist() == [
        [0, 0, 0],
        [0, 0, 0],
        [],
    ]
    assert (
        ak.str.count_substring_regex(bytestring_repeats, rb"BA\d*").layout.form
        == ak.str.count_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), rb"BA\d*"
        ).layout.form
    )

    assert ak.str.count_substring_regex(
        bytestring_repeats, rb"BA\d*", ignore_case=True
    ).tolist() == [
        [2, 0, 1],
        [0, 1, 1],
        [],
    ]
    assert (
        ak.str.count_substring_regex(
            bytestring_repeats, rb"BA\d*", ignore_case=True
        ).layout.form
        == ak.str.count_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), rb"BA\d*", ignore_case=True
        ).layout.form
    )

    assert ak.str.count_substring_regex(bytestring_repeats, rb"\d{1,}").tolist() == [
        [2, 0, 0],
        [1, 1, 1],
        [],
    ]
    assert (
        ak.str.count_substring_regex(bytestring_repeats, rb"\d{1,}").layout.form
        == ak.str.count_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), rb"\d{1,}"
        ).layout.form
    )


def test_ends_with():
    assert ak.str.ends_with(string_repeats, "BAR").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.ends_with(string_repeats, "BAR").layout.form
        == ak.str.ends_with(
            ak.to_backend(string_repeats, "typetracer"), "BAR"
        ).layout.form
    )

    assert ak.str.ends_with(string_repeats, "BAR", ignore_case=True).tolist() == [
        [False, False, True],
        [False, True, True],
        [],
    ]
    assert (
        ak.str.ends_with(string_repeats, "BAR", ignore_case=True).layout.form
        == ak.str.ends_with(
            ak.to_backend(string_repeats, "typetracer"), "BAR", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.ends_with(bytestring_repeats, b"BAR").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.ends_with(bytestring_repeats, b"BAR").layout.form
        == ak.str.ends_with(
            ak.to_backend(bytestring_repeats, "typetracer"), b"BAR"
        ).layout.form
    )

    assert ak.str.ends_with(bytestring_repeats, b"BAR", ignore_case=True).tolist() == [
        [False, False, True],
        [False, True, True],
        [],
    ]
    assert (
        ak.str.ends_with(bytestring_repeats, b"BAR", ignore_case=True).layout.form
        == ak.str.ends_with(
            ak.to_backend(bytestring_repeats, "typetracer"), b"BAR", ignore_case=True
        ).layout.form
    )


def test_starts_with():
    assert ak.str.starts_with(string_repeats, "FOO").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.starts_with(string_repeats, "FOO").layout.form
        == ak.str.starts_with(
            ak.to_backend(string_repeats, "typetracer"), "FOO"
        ).layout.form
    )

    assert ak.str.starts_with(string_repeats, "FOO", ignore_case=True).tolist() == [
        [True, True, False],
        [False, False, True],
        [],
    ]
    assert (
        ak.str.starts_with(string_repeats, "FOO", ignore_case=True).layout.form
        == ak.str.starts_with(
            ak.to_backend(string_repeats, "typetracer"), "FOO", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.starts_with(bytestring_repeats, b"FOO").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.starts_with(bytestring_repeats, b"FOO").layout.form
        == ak.str.starts_with(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO"
        ).layout.form
    )

    assert ak.str.starts_with(
        bytestring_repeats, b"FOO", ignore_case=True
    ).tolist() == [
        [True, True, False],
        [False, False, True],
        [],
    ]
    assert (
        ak.str.starts_with(bytestring_repeats, b"FOO", ignore_case=True).layout.form
        == ak.str.starts_with(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO", ignore_case=True
        ).layout.form
    )


def test_find_substring():
    assert ak.str.find_substring(string_repeats, "FOO").tolist() == [
        [-1, -1, -1],
        [-1, -1, -1],
        [],
    ]
    assert (
        ak.str.find_substring(string_repeats, "FOO").layout.form
        == ak.str.find_substring(
            ak.to_backend(string_repeats, "typetracer"), "FOO"
        ).layout.form
    )

    assert ak.str.find_substring(string_repeats, "FOO", ignore_case=True).tolist() == [
        [0, 0, -1],
        [3, -1, 0],
        [],
    ]
    assert (
        ak.str.find_substring(string_repeats, "FOO", ignore_case=True).layout.form
        == ak.str.find_substring(
            ak.to_backend(string_repeats, "typetracer"), "FOO", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.find_substring(bytestring_repeats, b"FOO").tolist() == [
        [-1, -1, -1],
        [-1, -1, -1],
        [],
    ]
    assert (
        ak.str.find_substring(bytestring_repeats, b"FOO").layout.form
        == ak.str.find_substring(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO"
        ).layout.form
    )

    assert ak.str.find_substring(
        bytestring_repeats, b"FOO", ignore_case=True
    ).tolist() == [
        [0, 0, -1],
        [3, -1, 0],
        [],
    ]
    assert (
        ak.str.find_substring(bytestring_repeats, b"FOO", ignore_case=True).layout.form
        == ak.str.find_substring(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO", ignore_case=True
        ).layout.form
    )


def test_find_substring_regex():
    assert ak.str.find_substring_regex(string_repeats, r"FOO\d+").tolist() == [
        [-1, -1, -1],
        [-1, -1, -1],
        [],
    ]
    assert (
        ak.str.find_substring_regex(string_repeats, r"FOO\d+").layout.form
        == ak.str.find_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"FOO\d+"
        ).layout.form
    )

    assert ak.str.find_substring_regex(
        string_repeats, r"FOO\d+", ignore_case=True
    ).tolist() == [
        [0, -1, -1],
        [-1, -1, 0],
        [],
    ]
    assert (
        ak.str.find_substring_regex(
            string_repeats, r"FOO\d+", ignore_case=True
        ).layout.form
        == ak.str.find_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"FOO\d+", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.find_substring_regex(bytestring_repeats, rb"FOO\d+").tolist() == [
        [-1, -1, -1],
        [-1, -1, -1],
        [],
    ]
    assert (
        ak.str.find_substring_regex(bytestring_repeats, rb"FOO\d+").layout.form
        == ak.str.find_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), rb"FOO\d+"
        ).layout.form
    )

    assert ak.str.find_substring_regex(
        bytestring_repeats, rb"FOO\d+", ignore_case=True
    ).tolist() == [
        [0, -1, -1],
        [-1, -1, 0],
        [],
    ]
    assert (
        ak.str.find_substring_regex(
            bytestring_repeats, rb"FOO\d+", ignore_case=True
        ).layout.form
        == ak.str.find_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"),
            rb"FOO\d+",
            ignore_case=True,
        ).layout.form
    )


def test_match_like():
    assert ak.str.match_like(string_repeats, "FOO%BA%").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.match_like(string_repeats, "FOO%BA%").layout.form
        == ak.str.match_like(
            ak.to_backend(string_repeats, "typetracer"), "FOO%BA%"
        ).layout.form
    )

    assert ak.str.match_like(string_repeats, "FOO%BA%", ignore_case=True).tolist() == [
        [True, False, False],
        [False, False, True],
        [],
    ]
    assert (
        ak.str.match_like(string_repeats, "FOO%BA%", ignore_case=True).layout.form
        == ak.str.match_like(
            ak.to_backend(string_repeats, "typetracer"), "FOO%BA%", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.match_like(bytestring_repeats, b"FOO%BA%").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.match_like(bytestring_repeats, b"FOO%BA%").layout.form
        == ak.str.match_like(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO%BA%"
        ).layout.form
    )

    assert ak.str.match_like(
        bytestring_repeats, b"FOO%BA%", ignore_case=True
    ).tolist() == [
        [True, False, False],
        [False, False, True],
        [],
    ]
    assert (
        ak.str.match_like(bytestring_repeats, b"FOO%BA^", ignore_case=True).layout.form
        == ak.str.match_like(
            ak.to_backend(bytestring_repeats, "typetracer"),
            b"FOO%BA^",
            ignore_case=True,
        ).layout.form
    )


def test_match_substring():
    assert ak.str.match_substring(string_repeats, "FOO").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.match_substring(string_repeats, "FOO").layout.form
        == ak.str.match_substring(
            ak.to_backend(string_repeats, "typetracer"), "FOO"
        ).layout.form
    )

    assert ak.str.match_substring(string_repeats, "FOO", ignore_case=True).tolist() == [
        [True, True, False],
        [True, False, True],
        [],
    ]
    assert (
        ak.str.match_substring(string_repeats, "FOO", ignore_case=True).layout.form
        == ak.str.match_substring(
            ak.to_backend(string_repeats, "typetracer"), "FOO", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.match_substring(bytestring_repeats, b"FOO").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.match_substring(bytestring_repeats, b"FOO").layout.form
        == ak.str.match_substring(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO"
        ).layout.form
    )

    assert ak.str.match_substring(
        bytestring_repeats, b"FOO", ignore_case=True
    ).tolist() == [
        [True, True, False],
        [True, False, True],
        [],
    ]
    assert (
        ak.str.match_substring(bytestring_repeats, b"FOO", ignore_case=True).layout.form
        == ak.str.match_substring(
            ak.to_backend(bytestring_repeats, "typetracer"), b"FOO", ignore_case=True
        ).layout.form
    )


def test_match_substring_regex():
    assert ak.str.match_substring_regex(string_repeats, r"FOO\d+").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.match_substring_regex(string_repeats, r"FOO\d+").layout.form
        == ak.str.match_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"FOO\d+"
        ).layout.form
    )

    assert ak.str.match_substring_regex(
        string_repeats, r"FOO\d+", ignore_case=True
    ).tolist() == [
        [True, False, False],
        [False, False, True],
        [],
    ]
    assert (
        ak.str.match_substring_regex(
            string_repeats, r"FOO\d+", ignore_case=True
        ).layout.form
        == ak.str.match_substring_regex(
            ak.to_backend(string_repeats, "typetracer"), r"FOO\d+", ignore_case=True
        ).layout.form
    )

    # Bytestrings
    assert ak.str.match_substring_regex(bytestring_repeats, rb"FOO\d+").tolist() == [
        [False, False, False],
        [False, False, False],
        [],
    ]
    assert (
        ak.str.match_substring_regex(bytestring_repeats, rb"FOO\d+").layout.form
        == ak.str.match_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"), rb"FOO\d+"
        ).layout.form
    )

    assert ak.str.match_substring_regex(
        bytestring_repeats, rb"FOO\d+", ignore_case=True
    ).tolist() == [
        [True, False, False],
        [False, False, True],
        [],
    ]
    assert (
        ak.str.match_substring_regex(
            bytestring_repeats, rb"FOO\d+", ignore_case=True
        ).layout.form
        == ak.str.match_substring_regex(
            ak.to_backend(bytestring_repeats, "typetracer"),
            rb"FOO\d+",
            ignore_case=True,
        ).layout.form
    )


def test_is_in():
    assert ak.str.is_in(string_repeats, ["123foo", "foo"]).tolist() == [
        [False, True, False],
        [True, False, False],
        [],
    ]
    assert (
        ak.str.is_in(string_repeats, ["123foo", "foo"]).layout.form
        == ak.str.is_in(
            ak.to_backend(string_repeats, "typetracer"), ["123foo", "foo"]
        ).layout.form
    )

    assert ak.str.is_in(
        [
            ["foo123bar123baz", "foo", "bar"],
            ["123foo", "456bar", "foo123456bar"],
            [None],
        ],
        ["123foo", "foo", None],
    ).tolist() == [
        [False, True, False],
        [True, False, False],
        [True],
    ]
    assert (
        ak.str.is_in(
            [
                ["foo123bar123baz", "foo", "bar"],
                ["123foo", "456bar", "foo123456bar"],
                [None],
            ],
            ["123foo", "foo", None],
        ).layout.form
        == ak.str.is_in(
            ak.to_backend(
                [
                    ["foo123bar123baz", "foo", "bar"],
                    ["123foo", "456bar", "foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            ["123foo", "foo", None],
        ).layout.form
    )

    assert ak.str.is_in(
        [
            ["foo123bar123baz", "foo", "bar"],
            ["123foo", "456bar", "foo123456bar"],
            [None],
        ],
        ["123foo", "foo", None],
        skip_nones=True,
    ).tolist() == [
        [False, True, False],
        [True, False, False],
        [False],
    ]
    assert (
        ak.str.is_in(
            [
                ["foo123bar123baz", "foo", "bar"],
                ["123foo", "456bar", "foo123456bar"],
                [None],
            ],
            ["123foo", "foo", None],
            skip_nones=True,
        ).layout.form
        == ak.str.is_in(
            ak.to_backend(
                [
                    ["foo123bar123baz", "foo", "bar"],
                    ["123foo", "456bar", "foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            ["123foo", "foo", None],
            skip_nones=True,
        ).layout.form
    )

    # Bytestrings
    assert ak.str.is_in(bytestring_repeats, [b"123foo", b"foo"]).tolist() == [
        [False, True, False],
        [True, False, False],
        [],
    ]
    assert (
        ak.str.is_in(bytestring_repeats, [b"123foo", b"foo"]).layout.form
        == ak.str.is_in(
            ak.to_backend(bytestring_repeats, "typetracer"), [b"123foo", b"foo"]
        ).layout.form
    )

    assert ak.str.is_in(
        [
            [b"foo123bar123baz", b"foo", b"bar"],
            [b"123foo", b"456bar", b"foo123456bar"],
            [None],
        ],
        [b"123foo", b"foo", None],
    ).tolist() == [
        [False, True, False],
        [True, False, False],
        [True],
    ]
    assert (
        ak.str.is_in(
            [
                [b"foo123bar123baz", b"foo", b"bar"],
                [b"123foo", b"456bar", b"foo123456bar"],
                [None],
            ],
            [b"123foo", b"foo", None],
        ).layout.form
        == ak.str.is_in(
            ak.to_backend(
                [
                    [b"foo123bar123baz", b"foo", b"bar"],
                    [b"123foo", b"456bar", b"foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            [b"123foo", b"foo", None],
        ).layout.form
    )

    assert ak.str.is_in(
        [
            [b"foo123bar123baz", b"foo", b"bar"],
            [b"123foo", b"456bar", b"foo123456bar"],
            [None],
        ],
        [b"123foo", b"foo", None],
        skip_nones=True,
    ).tolist() == [
        [False, True, False],
        [True, False, False],
        [False],
    ]
    assert (
        ak.str.is_in(
            [
                [b"foo123bar123baz", b"foo", b"bar"],
                [b"123foo", b"456bar", b"foo123456bar"],
                [None],
            ],
            [b"123foo", b"foo", None],
            skip_nones=True,
        ).layout.form
        == ak.str.is_in(
            ak.to_backend(
                [
                    [b"foo123bar123baz", b"foo", b"bar"],
                    [b"123foo", b"456bar", b"foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            [b"123foo", b"foo", None],
            skip_nones=True,
        ).layout.form
    )


def test_index_in():
    assert ak.str.index_in(string_repeats, ["123foo", "foo"]).tolist() == [
        [None, 1, None],
        [0, None, None],
        [],
    ]
    assert (
        ak.str.index_in(string_repeats, ["123foo", "foo"]).layout.form
        == ak.str.index_in(
            ak.to_backend(string_repeats, "typetracer"), ["123foo", "foo"]
        ).layout.form
    )

    assert ak.str.index_in(
        [
            ["foo123bar123baz", "foo", "bar"],
            ["123foo", "456bar", "foo123456bar"],
            [None],
        ],
        ["123foo", "foo", None],
    ).tolist() == [
        [None, 1, None],
        [0, None, None],
        [2],
    ]
    assert (
        ak.str.index_in(
            [
                ["foo123bar123baz", "foo", "bar"],
                ["123foo", "456bar", "foo123456bar"],
                [None],
            ],
            ["123foo", "foo", None],
        ).layout.form
        == ak.str.index_in(
            ak.to_backend(
                [
                    ["foo123bar123baz", "foo", "bar"],
                    ["123foo", "456bar", "foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            ["123foo", "foo", None],
        ).layout.form
    )

    assert ak.str.index_in(
        [
            ["foo123bar123baz", "foo", "bar"],
            ["123foo", "456bar", "foo123456bar"],
            [None],
        ],
        ["123foo", "foo", None],
        skip_nones=True,
    ).tolist() == [
        [None, 1, None],
        [0, None, None],
        [None],
    ]
    assert (
        ak.str.index_in(
            [
                ["foo123bar123baz", "foo", "bar"],
                ["123foo", "456bar", "foo123456bar"],
                [None],
            ],
            ["123foo", "foo", None],
            skip_nones=True,
        ).layout.form
        == ak.str.index_in(
            ak.to_backend(
                [
                    ["foo123bar123baz", "foo", "bar"],
                    ["123foo", "456bar", "foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            ["123foo", "foo", None],
            skip_nones=True,
        ).layout.form
    )

    # Bytestrings
    assert ak.str.index_in(string_repeats, [b"123foo", b"foo"]).tolist() == [
        [None, 1, None],
        [0, None, None],
        [],
    ]
    assert (
        ak.str.index_in(bytestring_repeats, [b"123foo", b"foo"]).layout.form
        == ak.str.index_in(
            ak.to_backend(bytestring_repeats, "typetracer"), [b"123foo", b"foo"]
        ).layout.form
    )

    assert ak.str.index_in(
        [
            [b"foo123bar123baz", b"foo", b"bar"],
            [b"123foo", b"456bar", b"foo123456bar"],
            [None],
        ],
        [b"123foo", b"foo", None],
    ).tolist() == [
        [None, 1, None],
        [0, None, None],
        [2],
    ]
    assert (
        ak.str.index_in(
            [
                [b"foo123bar123baz", b"foo", b"bar"],
                [b"123foo", b"456bar", b"foo123456bar"],
                [None],
            ],
            [b"123foo", b"foo", None],
        ).layout.form
        == ak.str.index_in(
            ak.to_backend(
                [
                    [b"foo123bar123baz", b"foo", b"bar"],
                    [b"123foo", b"456bar", b"foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            [b"123foo", b"foo", None],
        ).layout.form
    )

    assert ak.str.index_in(
        [
            [b"foo123bar123baz", b"foo", b"bar"],
            [b"123foo", b"456bar", b"foo123456bar"],
            [None],
        ],
        [b"123foo", b"foo", None],
        skip_nones=True,
    ).tolist() == [
        [None, 1, None],
        [0, None, None],
        [None],
    ]
    assert (
        ak.str.index_in(
            [
                [b"foo123bar123baz", b"foo", b"bar"],
                [b"123foo", b"456bar", b"foo123456bar"],
                [None],
            ],
            [b"123foo", b"foo", None],
            skip_nones=True,
        ).layout.form
        == ak.str.index_in(
            ak.to_backend(
                [
                    [b"foo123bar123baz", b"foo", b"bar"],
                    [b"123foo", b"456bar", b"foo123456bar"],
                    [None],
                ],
                "typetracer",
            ),
            [b"123foo", b"foo", None],
            skip_nones=True,
        ).layout.form
    )


def test_to_categorical():
    assert (
        ak.str.to_categorical(["foo", "bar", "bar", "fee"]).layout.form
        == ak.str.to_categorical(
            ak.to_backend(["foo", "bar", "bar", "fee"], "typetracer")
        ).layout.form
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from abc import abstractmethod

import numpy

from awkward._nplikes.array_like import ArrayLike
from awkward._nplikes.shape import ShapeItem
from awkward._singleton import PublicSingleton
from awkward._typing import (
    TYPE_CHECKING,
    Any,
    DType,
    Literal,
    NamedTuple,
    Protocol,
    TypeAlias,
    TypeVar,
)

if TYPE_CHECKING:
    from numpy.typing import DTypeLike

    from awkward._nplikes.placeholder import PlaceholderArray


IndexType: TypeAlias = "int | ArrayLikeT"


class UniqueAllResult(NamedTuple):
    values: ArrayLike
    indices: ArrayLike
    inverse_indices: ArrayLike
    counts: ArrayLike


class NumpyMetadata(PublicSingleton):
    bool_ = numpy.bool_
    int8 = numpy.int8
    int16 = numpy.int16
    int32 = numpy.int32
    int64 = numpy.int64
    uint8 = numpy.uint8
    uint16 = numpy.uint16
    uint32 = numpy.uint32
    uint64 = numpy.uint64
    longlong = numpy.longlong
    float32 = numpy.float32
    float64 = numpy.float64
    complex64 = numpy.complex64
    complex128 = numpy.complex128
    str_ = numpy.str_
    bytes_ = numpy.bytes_

    datetime64 = numpy.datetime64
    timedelta64 = numpy.timedelta64

    intp = numpy.intp
    integer = numpy.integer
    signedinteger = numpy.signedinteger
    unsignedinteger = numpy.unsignedinteger
    floating = numpy.floating
    complexfloating = numpy.complexfloating
    number = numpy.number
    object_ = numpy.object_
    generic = numpy.generic

    dtype = numpy.dtype
    ufunc = numpy.ufunc
    iinfo = numpy.iinfo
    finfo = numpy.finfo
    errstate = numpy.errstate
    newaxis: None = numpy.newaxis

    ndarray = numpy.ndarray

    nan = numpy.nan
    inf = numpy.inf

    nat = numpy.datetime64("NaT")
    datetime_data = staticmethod(numpy.datetime_data)
    issubdtype = staticmethod(numpy.issubdtype)

    AxisError = numpy.AxisError


if hasattr(numpy, "float16"):
    NumpyMetadata.float16 = numpy.float16  # type: ignore[attr-defined]

if hasattr(numpy, "float128"):
    NumpyMetadata.float128 = numpy.float128  # type: ignore[attr-defined]

if hasattr(numpy, "complex256"):
    NumpyMetadata.complex256 = numpy.complex256  # type: ignore[attr-defined]


ArrayLikeT = TypeVar("ArrayLikeT", bound=ArrayLike)


class UfuncLike(Protocol):
    nargs: int
    nin: int
    nout: int

    def resolve_dtypes(
        self, dtypes: tuple[DType | type[int] | type[complex] | type[float] | None, ...]
    ) -> tuple[DType, ...]:
        ...

    def __call__(self, *args: ArrayLikeT, **kwargs) -> ArrayLikeT:
        ...


class NumpyLike(PublicSingleton, Protocol[ArrayLikeT]):
    ############################ Awkward features

    @abstractmethod
    def apply_ufunc(
        self,
        ufunc: UfuncLike,
        method: str,
        args: list[Any],
        kwargs: dict[str, Any] | None = None,
    ) -> ArrayLikeT | tuple[ArrayLikeT, ...]:
        ...

    @property
    @abstractmethod
    def supports_structured_dtypes(self) -> bool:
        ...

    @property
    @abstractmethod
    def known_data(self) -> bool:
        ...

    @property
    @abstractmethod
    def is_eager(self) -> bool:
        ...

    ############################ array creation

    @abstractmethod
    def asarray(
        self,
        obj,
        *,
        dtype: DTypeLike | None = None,
        copy: bool | None = None,
    ) -> ArrayLikeT | PlaceholderArray:
        ...

    # FIXME: find a way to express TypeVar(..., OtherTypeVar(...), FOO) such that
    #        this function preserves the type identity of the input
    @abstractmethod
    def ascontiguousarray(
        self, x: ArrayLikeT | PlaceholderArray
    ) -> ArrayLikeT | PlaceholderArray:
        ...

    @abstractmethod
    def frombuffer(
        self, buffer, *, dtype: DTypeLike | None = None, count: ShapeItem = -1
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def from_dlpack(self, x: Any) -> ArrayLikeT:
        ...

    @abstractmethod
    def zeros(
        self,
        shape: ShapeItem | tuple[ShapeItem, ...],
        *,
        dtype: DTypeLike | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def ones(
        self,
        shape: ShapeItem | tuple[ShapeItem, ...],
        *,
        dtype: DTypeLike | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def empty(
        self,
        shape: ShapeItem | tuple[ShapeItem, ...],
        *,
        dtype: DTypeLike | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def full(
        self,
        shape: ShapeItem | tuple[ShapeItem, ...],
        fill_value,
        *,
        dtype: DTypeLike | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def zeros_like(
        self, x: ArrayLikeT | PlaceholderArray, *, dtype: DTypeLike | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def ones_like(
        self, x: ArrayLikeT | PlaceholderArray, *, dtype: DTypeLike | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def full_like(
        self,
        x: ArrayLikeT | PlaceholderArray,
        fill_value,
        *,
        dtype: DTypeLike | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def arange(
        self,
        start: float | int,
        stop: float | int | None = None,
        step: float | int = 1,
        *,
        dtype: DTypeLike | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def meshgrid(
        self, *arrays: ArrayLikeT, indexing: Literal["xy", "ij"] = "xy"
    ) -> list[ArrayLikeT]:
        ...

    ############################ testing

    @abstractmethod
    def array_equal(
        self, x1: ArrayLikeT, x2: ArrayLikeT, *, equal_nan: bool = False
    ) -> bool:
        ...

    @abstractmethod
    def searchsorted(
        self,
        x: ArrayLikeT,
        values: ArrayLikeT,
        *,
        side: Literal["left", "right"] = "left",
        sorter: ArrayLikeT | None = None,
    ) -> ArrayLikeT:
        ...

    ############################ manipulation

    @abstractmethod
    def broadcast_arrays(self, *arrays: ArrayLikeT) -> list[ArrayLikeT]:
        ...

    @abstractmethod
    def broadcast_to(self, x: ArrayLikeT, shape: tuple[ShapeItem, ...]) -> ArrayLikeT:
        ...

    @abstractmethod
    def shape_item_as_index(self, x1: ShapeItem) -> int | ArrayLikeT:
        ...

    @abstractmethod
    def index_as_shape_item(self, x1: IndexType) -> ShapeItem:
        ...

    @abstractmethod
    def derive_slice_for_length(
        self, slice_: slice, length: ShapeItem
    ) -> tuple[IndexType, IndexType, IndexType, ShapeItem]:
        ...

    @abstractmethod
    def regularize_index_for_length(
        self, index: IndexType, length: ShapeItem
    ) -> IndexType:
        ...

    # FIXME: find a way to express TypeVar(..., OtherTypeVar(...), FOO) such that
    #        this function preserves the type identity of the input
    @abstractmethod
    def reshape(
        self,
        x: ArrayLikeT | PlaceholderArray,
        shape: tuple[ShapeItem, ...],
        *,
        copy: bool | None = None,
    ) -> ArrayLikeT | PlaceholderArray:
        ...

    @abstractmethod
    def nonzero(self, x: ArrayLikeT) -> tuple[ArrayLikeT, ...]:
        ...

    @abstractmethod
    def where(
        self, condition: ArrayLikeT, x1: ArrayLikeT, x2: ArrayLikeT
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def unique_values(self, x: ArrayLikeT) -> ArrayLikeT:
        ...

    @abstractmethod
    def unique_all(self, x: ArrayLikeT) -> UniqueAllResult:
        ...

    @abstractmethod
    def sort(
        self,
        x: ArrayLikeT,
        *,
        axis: int = -1,
        descending: bool = False,
        stable: bool = True,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def concat(
        self,
        arrays: list[ArrayLikeT] | tuple[ArrayLikeT, ...],
        *,
        axis: int | None = 0,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def repeat(
        self,
        x: ArrayLikeT,
        repeats: ArrayLikeT | int,
        *,
        axis: int | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def stack(
        self,
        arrays: list[ArrayLikeT] | tuple[ArrayLikeT, ...],
        *,
        axis: int = 0,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def packbits(
        self,
        x: ArrayLikeT,
        *,
        axis: int | None = None,
        bitorder: Literal["big", "little"] = "big",
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def unpackbits(
        self,
        x: ArrayLikeT,
        *,
        axis: int | None = None,
        count: int | None = None,
        bitorder: Literal["big", "little"] = "big",
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def strides(self, x: ArrayLikeT | PlaceholderArray) -> tuple[ShapeItem, ...]:
        ...

    ############################ ufuncs

    @abstractmethod
    def add(
        self, x1: ArrayLikeT, x2: ArrayLikeT, maybe_out: ArrayLikeT | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def logical_or(
        self, x1: ArrayLikeT, x2: ArrayLikeT, *, maybe_out: ArrayLikeT | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def logical_and(
        self, x1: ArrayLikeT, x2: ArrayLikeT, *, maybe_out: ArrayLikeT | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def logical_not(
        self, x: ArrayLikeT, maybe_out: ArrayLikeT | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def sqrt(self, x: ArrayLikeT, maybe_out: ArrayLikeT | None = None) -> ArrayLikeT:
        ...

    @abstractmethod
    def exp(self, x: ArrayLikeT, maybe_out: ArrayLikeT | None = None) -> ArrayLikeT:
        ...

    @abstractmethod
    def divide(
        self, x1: ArrayLikeT, x2: ArrayLikeT, maybe_out: ArrayLikeT | None = None
    ) -> ArrayLikeT:
        ...

    ############################ almost-ufuncs

    @abstractmethod
    def nan_to_num(
        self,
        x: ArrayLikeT,
        *,
        copy: bool = True,
        nan: int | float | None = 0.0,
        posinf: int | float | None = None,
        neginf: int | float | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def isclose(
        self,
        x1: ArrayLikeT,
        x2: ArrayLikeT,
        *,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        equal_nan: bool = False,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def isnan(self, x: ArrayLikeT) -> ArrayLikeT:
        ...

    @abstractmethod
    def all(
        self,
        x: ArrayLikeT,
        *,
        axis: int | tuple[int, ...] | None = None,
        keepdims: bool = False,
        maybe_out: ArrayLikeT | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def any(
        self,
        x: ArrayLikeT,
        *,
        axis: int | tuple[int, ...] | None = None,
        keepdims: bool = False,
        maybe_out: ArrayLikeT | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def min(
        self,
        x: ArrayLikeT,
        *,
        axis: int | tuple[int, ...] | None = None,
        keepdims: bool = False,
        maybe_out: ArrayLikeT | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def max(
        self,
        x: ArrayLikeT,
        *,
        axis: int | tuple[int, ...] | None = None,
        keepdims: bool = False,
        maybe_out: ArrayLikeT | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def count_nonzero(
        self, x: ArrayLikeT, *, axis: int | tuple[int, ...] | None = None
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def cumsum(
        self,
        x: ArrayLikeT,
        *,
        axis: int | None = None,
        maybe_out: ArrayLikeT | None = None,
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def array_str(
        self,
        x: ArrayLikeT,
        *,
        max_line_width: int | None = None,
        precision: int | None = None,
        suppress_small: bool | None = None,
    ):
        ...

    @abstractmethod
    def astype(
        self, x: ArrayLikeT, dtype: DTypeLike, *, copy: bool | None = True
    ) -> ArrayLikeT:
        ...

    @abstractmethod
    def can_cast(self, from_: DType | ArrayLikeT, to: DType | ArrayLikeT) -> bool:
        ...

    @abstractmethod
    def is_c_contiguous(self, x: ArrayLikeT | PlaceholderArray) -> bool:
        ...

    @classmethod
    @abstractmethod
    def is_own_array(cls, obj) -> bool:
        ...

    @classmethod
    @abstractmethod
    def is_own_array_type(cls, type_: type) -> bool:
        ...



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_UnionArray8_32_project_64_1():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_2():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 2, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_3():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 3, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_4():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 4, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_5():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_6():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    fromindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    length = 3
    which = 1
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_7():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    fromindex = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    length = 3
    which = 1
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 2, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_8():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    fromindex = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    length = 3
    which = 1
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 3, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_9():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    fromindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    length = 3
    which = 1
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 4, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_10():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    fromindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length = 3
    which = 1
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_11():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_12():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 2, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_13():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 3, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_14():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [1, 4, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_UnionArray8_32_project_64_15():
    lenout = cupy.array([123], dtype=cupy.int64)
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    fromindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length = 3
    which = 0
    funcC = cupy_backend['awkward_UnionArray_project', cupy.int64, cupy.int64, cupy.int8, cupy.int32]
    funcC(lenout, tocarry, fromtags, fromindex, length, which)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_lenout = [3]
    assert cupy.array_equal(lenout[:len(pytest_lenout)], cupy.array(pytest_lenout))
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy

__all__ = ("FieldNotFoundError", "AxisError")


class FieldNotFoundError(IndexError):
    ...


AxisError = numpy.AxisError



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_RegularArray_getitem_next_at_64_1():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    at = 0
    length = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_at', cupy.int64]
    funcC(tocarry, at, length, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 3, 6]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_RegularArray_getitem_next_at_64_2():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    at = 0
    length = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_at', cupy.int64]
    funcC(tocarry, at, length, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 3, 6]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_RegularArray_getitem_next_at_64_3():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    at = 2
    length = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_at', cupy.int64]
    funcC(tocarry, at, length, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [2, 5, 8]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_RegularArray_getitem_next_at_64_4():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    at = 1
    length = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_at', cupy.int64]
    funcC(tocarry, at, length, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 4, 7]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("slice",)


@high_level_function(module="ak.str")
def slice(
    array, start, stop=None, step=1, *, highlevel=True, behavior=None, attrs=None
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        start (int): Index to start slicing at (inclusive).
        stop (None or int): Index to stop slicing at (exclusive). If not given,
            slicing will stop at the end.
        step (int): Slice step.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Replaces any string or bytestring-valued data with a slice between `start`
    and `stop` indexes; `start` is inclusive and `stop` is exclusive and both
    are 0-indexed.

    For strings, `start` and `stop` are measured in Unicode characters; for
    bytestrings, `start` and `stop` are measured in bytes.

    The `start`, `stop`, and `replacement` are scalars; they cannot be
    different for each string/bytestring in the sample.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.utf8_slice_codeunits](https://arrow.apache.org/docs/python/generated/pyarrow.compute.utf8_slice_codeunits.html)
    or performs a literal slice on strings and bytestrings, respectively.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, start, stop, step, highlevel, behavior, attrs)


def _impl(array, start, stop, step, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute
    from awkward.operations.str import _apply_through_arrow

    pc = import_pyarrow_compute("ak.str.slice")

    def action(layout, **absorb):
        if layout.is_list and layout.parameter("__array__") == "string":
            return _apply_through_arrow(
                pc.utf8_slice_codeunits, layout, start, stop, step
            )

        elif layout.is_list and layout.parameter("__array__") == "bytestring":
            return layout[:, start:stop:step]

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(
            array,
            allow_record=False,
            allow_unknown=False,
            primitive_policy="error",
            string_policy="as-characters",
        )
    out = ak._do.recursively_apply(layout, action)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.typetracer import TypeTracer

typetracer = TypeTracer.instance()


def test():
    buffer = b"Hello world, what's the news?"
    left = np.frombuffer(buffer, np.float32, 4)
    right = typetracer.frombuffer(buffer, dtype=np.float32, count=4)
    assert left.dtype == right.dtype
    assert left.shape == right.shape


def test_zeros_like_numeric():
    source = ak.Array([[1, 2, 3]], backend="typetracer")
    result = ak.zeros_like(source)

    # Known lengths and dtypes should match
    assert result.layout.content.dtype == source.layout.content.dtype
    assert result.layout.content.length == source.layout.content.length


def test_zeros_like_strings():
    # Check strings!
    source = ak.Array(["hello", "world"], backend="typetracer")
    result = ak.zeros_like(source)
    assert result.layout.content.dtype == source.layout.content.dtype
    # zeros_like for strings means empty-string!
    assert result.layout.content.length == 0



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._attrs import attrs_of_obj
from awkward._backends.dispatch import backend_of
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of_obj
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout

__all__ = ("broadcast_fields",)

cpu = NumpyBackend.instance()


@high_level_function()
def broadcast_fields(*arrays, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        arrays: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Return a list of arrays whose types contain the same number of fields. Unlike
    #ak.broadcast_arrays, this function does not require record types to occur at the
    same depths. Where fields are missing from one record, they are inserted at the same
    position with an `option[unknown]` type. This type is easily erased by ufunc and
    concatenation operations.

        >>> x, y = ak.broadcast_fields(
        ...     [{"x": {"y": 1, "z": 2, "w": [1]}}],
        ...     [{"x": [{"y": 1}]}],
        ... )
        >>> x.type.show()
        1 * {
            x: {
                y: int64,
                z: int64,
                w: var * int64
            }
        }
        >>> y.type.show()
        1 * {
            x: var * {
                y: int64,
                z: ?unknown,
                w: ?unknown
            }
        }

    """
    # Dispatch
    yield arrays

    # Implementation
    return _impl(arrays, highlevel, behavior, attrs)


def _identity(content):
    return content


# A "pull-back" is a function that takes a leaf-node, and rebuilds the
# tree between the leaf and the caller. i.e., for some path X.Y.Z.LEAF,
# this function returns (f, LEAF) such that f(LEAF) = X.Y.Z.LEAF.
# Using pull-backs rather than recursive descent allows for the control
# flow to be implemented at the call-site rather than the leaves.
def _descend_to_record_or_leaf(layout, pullback=_identity):
    assert layout is not None
    if isinstance(layout, ak.record.Record):
        return _descend_to_record_or_leaf(
            layout.array, lambda x: ak.record.Record(x, layout.at)
        )
    elif layout.is_record or layout.is_identity_like or layout.is_leaf:
        return pullback, layout
    elif layout.is_option or layout.is_indexed or layout.is_list:

        def next_pull(content):
            return pullback(layout.copy(content=content))

        return _descend_to_record_or_leaf(layout.content, next_pull)
    elif layout.is_union:
        raise TypeError("unions are not supported")
    else:
        raise AssertionError("unexpected content type")


# Like broadcast_and_apply, we want to walk into each layout, correct the structure, and then rebuilt the arrays
# We do this using "pull back" functions that accept a child content, and return the top-level layout. Unlike
# layout.copy, the pull-back functions can be arbitrarily deep: the closures maintain the structure of the array
def _recurse(inputs):
    # Descend to records, identities, or leaves
    pullbacks, next_inputs = zip(*[_descend_to_record_or_leaf(x) for x in inputs])
    # With no records, we can exit here
    if not any(c.is_record for c in next_inputs):
        return [pull(layout) for pull, layout in zip(pullbacks, next_inputs)]
    # Otherwise, we can only work with all non-record, or all record/identity
    elif not all(c.is_record or c.is_identity_like for c in next_inputs):
        raise AssertionError(
            "if any inputs are records, all inputs must be records or identities"
        )

    # Broadcast the fields of only the records
    next_records = [r for r in next_inputs if r.is_record]
    all_fields = ak._util.unique_list(
        [f for layout in next_records for f in layout.fields]
    )

    # Build a list of layouts for each field, i.e. [{x: aaaa, y: aaaa}, {x: bbbb, y: bbbb}] becomes
    # [[aaaa, bbbb], [aaaa, bbbb]], where fields = [x, y]
    # These layouts will be "broadcast" against each other, hence the per-field ordering
    layouts_by_field = []
    for field in all_fields:
        layouts_to_recurse = []
        layouts_for_field = []
        for layout in next_records:
            if layout.has_field(field):
                layouts_for_field.append(None)
                layouts_to_recurse.append(layout.content(field))
            else:
                layouts_for_field.append(layout.maybe_content(field))

        # We only want to recurse into non-missing fields, so we build this list separately as a generator
        recursed_field_layouts = iter(_recurse(layouts_to_recurse))

        # Now we build the final list of layouts for this field, choosing between the recursion result and the
        # original layout according to whether the layout was recursed into
        # The pattern here is that `layouts_for_field` maintains positional correspondence with the `records`,
        # but uses `None` as a token for "recursed". In this case, we take the layout from `recursed_field_layouts`
        # using the knowledge that `len(layouts_to_recurse)` corresponds to the number of `None`s
        layouts_by_field.append(
            [
                next(recursed_field_layouts) if layout is None else layout
                for layout in layouts_for_field
            ]
        )

    # Now we transpose the list-of-lists to group layouts by original record, instead of by the field
    layouts_by_record = zip(*layouts_by_field)
    # Rebuild the original records with the new fields
    next_records = iter(
        [
            record.copy(
                fields=all_fields,
                contents=contents,
            )
            for record, contents in zip(next_records, layouts_by_record)
        ]
    )

    # Merge the records and identities
    inner_layouts = [
        (layout if layout.is_identity_like else next(next_records))
        for layout in next_inputs
    ]

    # Rebuild the outermost layouts using pull-back functions
    return [pull(layout) for pull, layout in zip(pullbacks, inner_layouts)]


def _impl(arrays, highlevel, behavior, attrs):
    # Need at least one array!
    if len(arrays) == 0:
        return []

    backend = backend_of(*arrays, default=cpu)
    layouts = [ak.to_layout(x, allow_record=True).to_backend(backend) for x in arrays]

    result_layouts = _recurse(
        [
            record.array[record.at : record.at + 1]
            if isinstance(record, ak.record.Record)
            else record
            for record in layouts
        ]
    )

    return [
        wrap_layout(
            layout_out,
            behavior=behavior_of_obj(array_in, behavior=behavior),
            highlevel=highlevel,
            attrs=attrs_of_obj(array_in, attrs=attrs),
        )
        for layout_out, array_in in zip(result_layouts, arrays)
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest  # noqa: F401

import awkward as ak


def test_all_none():
    result = ak.argsort([None, None, None])
    assert ak.is_valid(result)
    assert result.to_list() == [0, 1, 2]
    assert isinstance(result.layout, ak.contents.NumpyArray)
    assert result.type == ak.types.ArrayType(ak.types.NumpyType("int64"), 3)


def test_mixed_none_local():
    result = ak.argsort([None, 1, None, 2, 0, None, -1])
    assert ak.is_valid(result)
    assert result.to_list() == [6, 4, 1, 3, 0, 2, 5]
    assert result.type == ak.types.ArrayType(ak.types.NumpyType("int64"), 7)


def test_mixed_none_2d_local():
    result = ak.argsort(
        [
            [None, 1, None, 0, None, None, -1],
            None,
            [None, 2, None, 2, 0, None, -2],
        ],
        axis=1,
    )
    assert ak.is_valid(result)
    assert result.to_list() == [[6, 3, 1, 0, 2, 4, 5], None, [6, 4, 1, 3, 0, 2, 5]]
    assert result.type == ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("int64"))), 3
    )


def test_mixed_none_2d_nonlocal():
    result = ak.argsort(
        [
            [None, 1, None, 0, None, None, -1],
            [None, 2, None, 2, 0, None, -2],
        ],
        axis=0,
    )
    assert ak.is_valid(result)
    assert result.to_list() == [[0, 0, 0, 0, 1, 0, 1], [1, 1, 1, 1, 0, 1, 0]]
    assert result.type == ak.types.ArrayType(
        ak.types.ListType(ak.types.NumpyType("int64")), 2
    )
